# Routines to compute correlations between samples of observables 
import numpy as np
import time


def auto_window(IATs, c):
    '''
    Windowing procedure of Caracciolo and Sokal, *Dynamic critical exponent of some Monte Carlo algorithms for the self-avoiding walk*, `J. Phys. A: Math. Gen. 19
    (1986) 797 <https://doi.org/10.1088/0305-4470/19/13/008>`_ to truncate the sum for the IAT.

    Parameters
    ----------
    IATs: array
        integrated autocorrelation time with increasingly late termination of the sum
    c: float
        defines the window width. For correlations that decay exponentially, a value of 4 or 5 is conventional

    Returns 
    -------
    idx: int
        index for array IATs, representing the IAT estimate from the windowing procedure
    ''' 
    ts = np.arange(len(IATs)) # all possible separation endpoints
    m =  ts < c * IATs # first occurrence where this is false gives IAT 
    if np.any(m):
        idx = np.argmin(m)
    else:
        idx = len(IATs) - 1
    return idx


def autocorr_func_1d(x):
    '''
    Computes the autocorrelation of a 1D array ``x`` using FFT and the `cross correlation theorem <https://mathworld.wolfram.com/Cross-CorrelationTheorem.html>`_.
    As FFTs yield circular convolutions and work most efficiently when the number of elements is a power of 2, pad the data with zeros to the next power of 2. 

    Parameters
    ----------
    x: (N,) array
        1 dimensional data series of N elements to find the autocorrelation function for

    Returns
    -------
    acf: (N,) array
        the autocorrelation function
    '''
    x = np.atleast_1d(x)
    if len(x.shape) != 1:
        raise ValueError("invalid dimensions for 1D autocorrelation function")

    def next_pow_two(n):
        i = 1
        while i < n:
            i = i << 1 # shifts bits to the left by one position i.e. multiplies by 2 
        return i
    
    n = next_pow_two(len(x))

    # Compute the FFT and then the auto-correlation function
    f = np.fft.fft(x - np.mean(x), n=2 * n)
    acf = np.fft.ifft(f * np.conjugate(f))[: len(x)].real
    acf /= 4 * n
    # normalize to get autocorrelation rather than autocovariance
    acf /= acf[0]

    return acf


def autocorrelator_repeats(data, c=4.0):
    '''
    Based on the implementation in `emcee <https://emcee.readthedocs.io/en/stable/tutorials/autocorr/>`_.
    Computes the autocorrelation function and integrated autocorrelation time (IAT) for passed ``data`` which is a 2D array such that each row represents a sample of observations.
    The correlations are computed between rows of the data by finding the 1D autocorrelation function for each column of the data. The overall ACF between rows is estimated
    as the average of correlations across columns. This also allows to get an estimate of the error of the ACF.
    An alternative is to average along rows first and to estimate the ACF as the autocorrelation of the final column (`Goodman, Weare 2010 <https://msp.org/camcos/2010/5-1/p04.xhtml>`_).

    Parameters
    ----------
    data: (M,L) array
        each row contains L samples of the same observable while different rows correspond to different positions in the chain of M observations.
        The correlation is computed across axis 1 i.e. gives the AFC for samples from different positions in the chain. 
    c: float, optional
        parameter used in determining when to truncate the sum for the IAT

    Returns
    -------
    ts: (M,) array
        array of considered separations between two samples
    ACF: (M,) array
        autocorrelation function
    ACF_err: (M,) array
        error of the autocorrelation function
    IAT: float
        integrated autocorrelation time, showing how many rows in the data lie between uncorrelated samples
    IAT_err: float
        error of the autocorrelation time
    delta_t: float
        time needed to compute the ACF and the IAT
    '''
    M, L = data.shape
    ts = np.arange(M)
    t1 = time.time()

    # get ACF and its error
    ACFs = np.zeros_like(data)
    for i in range(L):
        ACFs[:,i] = autocorr_func_1d(data[:,i])

    ACF, ACF_err = np.mean(ACFs, axis=1), np.std(ACFs, axis=1) / np.sqrt(L)

    # get all possible IAT and apply windowing
    IATs = 2.0 * np.cumsum(ACF) - 1.0 # IAT defined as 1 + sum starting from separation=1, but cumsum starts with t=0 for which ACF=1
    break_idx = auto_window(IATs, c)
    IAT = IATs[break_idx]
    IAT_err = np.sqrt((4*break_idx+2)/data.shape[0]) * IAT # Madras, Sokal 1988

    t2 = time.time()
    delta_t = t2-t1

    return ts, ACF, ACF_err, IAT, IAT_err, delta_t


def autocorrelator(data, c=4.0):
    '''
    Alias for :py:meth:`SU2xSU2.correlations.autocorrelator_repeats` when the data has been observed only once, meaning data is of shape (M,).

    See Also
    --------
    correlations.autocorrelator_repeats
    '''
    return autocorrelator_repeats(data.reshape((data.shape[0], 1)))


def corr_func_1D(x, y):
    '''
    Computes the correlation between the equal length 1D arrays ``x``, ``y`` using the cross correlation theorem.
    FFTs yield circular convolutions, such that ``x`` and ``y`` are assumed to have periodic boundary conditions. 
    When the data is not circular, need to pad it with zeros as done in :py:meth:`SU2xSU2.correlations.autocorr_func_1d`.

    Parameters
    ----------
    x: (N,) array
        first data series to correlate
    y: (N,) array
        second data series to correlate

    Returns
    -------
    cf: (N,) array
        correlation function between the data in x and y and of the same length as either array
    '''
    f = np.fft.fft(x)
    g = np.fft.fft(y)
    cf = np.fft.ifft(f * np.conjugate(g)).real

    return cf


def correlator_repeats(xs, ys):
    '''
    Computes the correlation function (CF) using FFTs between two equally sized 1D arrays for which several measurements exists. 
    Each column presents a new observation and correlations are computed along axis 0. The final CF is the average along axis 1.

    Parameters
    ----------
    xs: (L,M) array
        M measurements of first data vector of length L
    ys: (L,M) array
        M measurements of second data vector of length L

    Returns
    -------
    CF: (L,) array
        average correlation function based on the M measurements 
    CF_err: (L,) array
        uncertainty in the CF, estimated as the standard error on the mean corrected by the square root of the integrated autocorrelation time
    '''
    L, M = xs.shape # length of one data measurement, number of measurements

    # get CF and its naive error
    CFs = np.zeros_like(xs)
    for i in range(M):
        CFs[:,i] = corr_func_1D(xs[:,i], ys[:,i])

    CF, CF_err = np.mean(CFs, axis=1), np.std(CFs, axis=1) / np.sqrt(M)
    
    # correct CF error for autocorrelations
    for i in range(L):
        ts, ACF, ACF_err, IAT, IAT_err, delta_t = autocorrelator(CFs[i])
        CF_err[i] *= np.sqrt(IAT)

    return CF, CF_err


def correlator(xs, ys):
    '''
    Alias for :py:meth:`SU2xSU2.correlations.correlator_repeats` when the ``xs`` and ``ys`` have been observed only once, i.e. are both of shape (L,)
    This implies that the error of the correlation function cannot be estimated.

    See Also
    --------
    correlations.correlator_repeats
    '''
    return correlator_repeats(xs.reshape((xs.shape[0], 1)), ys.reshape((ys.shape[0], 1)))


# some additional plotting routines
import os
import numpy as np
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

# plt.style.use('scientific.mplstyle')
# plt.rcParams.update({'text.usetex': True}) # uncomment when latex is installed


def plot_chain(data_chain_path, ylabel, start_idx=0):
    '''
    Plots raw measurement chain against computer time to gauge when convergence has been achieved.
    Observables of particular interest are the susceptibility (as sufficient burn is needed for the critical slowing down analysis) and 
    the correlation function at a fixed separation (ideally chosen close to the correlation length as this is a slowly converging quantity
    and thus gives a lower bound for burn in).
    If the stored chain has already had some burn in removed, the starting value of the computer time can be adjusted by ``start_idx``. 

    Parameters
    ----------
    data_chain_path: str
        path to measurement chain file (must be .npy)
    ylabel: str
        label for the plotted measurement. When Latex commands are contained, pass as raw string i.e. r'$\chi$'
    start_idx: int (optional)
        amount of burn in already removed from the measurement chain
    '''
    data = np.load(data_chain_path)
    comp_time = np.arange(data.size) + start_idx

    fig = plt.figure()

    plt.plot(comp_time, data)
    plt.xlabel('computer time')
    plt.ylabel(ylabel)

    plt.show()


def correlation_func_plot(data_path, plot_path, fit_upper=None, show_plot=True, ybottom=None, xright=None):
    '''
    Produces a plot (with a logarithmic y axis) of the correlation function data (stored at ``data_path``) and the fitted, analytical expectation. 
    The plot is saved at ``plot_path`` with the fitting range and other plot parameters can be adjusted through the arguments.

    Allows manual adjustment of fitting the correlation length to the processed correlation function data (normalized and mirror averaged).
    A plot with the new fitting is produced and the inferred correlation length, its error and the associated chi2 are printed.
    These can then be manually added to (for example) a data/corlen_data.txt file.

    Parameters
    ----------
    data_path: str
        file path to the correlation function data that will be plotted. The file is assumed to be .npy and 
        contain the rows separation [0,L/2], correlation function, correlation function error respectively 
    plot_path: str 
        file path (including file extension) to store the plot at
    fit_upper: int (optional)
        largest separation (in units of the lattice spacing) to be included in the fit.
        If left as ``None``, only all non-zero values of the correlation function will be included in the fit
    show_plot: bool (optional)
        shows produced plot before saving it
    ybottom: float (optional)
        lower limit for y axis
    xright: float (optional)
        upper limit for x axis
    '''
    def fit(d,xi):
        return (np.cosh((d-L_2)/xi) - 1) / (np.cosh(L_2/xi) - 1)
    
    ds, cor, cor_err = np.load(data_path)
    L_2 = ds[-1]
    
    if fit_upper is not None:
        mask = ds <= fit_upper
    else:
        mask = cor > 0

    popt, pcov = curve_fit(fit, ds[mask], cor[mask], sigma=cor_err[mask], absolute_sigma=True)
    cor_length = popt[0] # in units of lattice spacing
    cor_length_err = np.sqrt(pcov[0][0])

    r = cor[mask] - fit(ds[mask], *popt)
    reduced_chi2 = np.sum((r/cor_err[mask])**2) / (mask.size - 1) # dof = number of observations - number of fitted parameters

    fig = plt.figure()

    plt.errorbar(ds, cor, yerr=cor_err, fmt='.', zorder=1)
    ds_fit = np.linspace(0, ds[mask][-1], 500)
    plt.plot(ds_fit, fit(ds_fit,*popt), c='g', zorder=2, label='$\\xi = %.3f \pm %.3f$\n $\chi_r^2 = %.2f$'%(cor_length, cor_length_err, reduced_chi2))
    # set axes limits
    if ybottom is not None:
        plt.ylim(bottom=ybottom, top=2)
    if xright is not None:
        plt.xlim(right=xright)
    plt.yscale('log')
    plt.xlabel(r'wall separation $d$ [$a$]')
    plt.ylabel(r'wall wall correlation $C_{ww}(d)$')
    plt.legend(loc='upper right')
    fig.gca().xaxis.set_major_locator(MaxNLocator(integer=True)) # set major ticks at integer positions only
    
    if show_plot:
        plt.show()

    # check if path contains directory. If it does, the directory is created should it not exist already
    dir_path = os.path.dirname(plot_path)
    if dir_path != '':
        os.makedirs(dir_path, exist_ok=True)
    fig.savefig(plot_path)
    plt.close() # for memory purposes

    print('corlen: %.18E \ncorlen_err: %.18E \nchi2: %.18E'%(cor_length, cor_length_err, reduced_chi2))
    return


def effective_mass_plot(data_path, xright=None, ytop=None, ybottom=None, show_plot=True, plot_path='plots/effective_mass.pdf'):
    '''
    Produces an effective mass plot using the processed correlation function data stored at ``data_path``.
    The data is expected to contain the separation, the correlation function value and its error row wise.
    The effective mass will be computed based on the assumption that the correlation function follows the shape of
    a cosh as analytically expected due to periodic boundary conditions. 
    As the effective mass becomes noisy for large separations, the plot range can be adjusted using the remaining keywords.  

    Parameters
    ----------
    data_path: str
        path to .npy file containing the averaged (ensemble and across the symmetry axis at L/2) and normalized correlation function
    xright: float
        upper limit of the separation shown in the plot
    ytop: float
        upper limit for the effective mass
    ybottom: float
        lower limit for the effective mass 
    '''

    def effective_mass(cor, cor_err):
        '''
        Computes the effective mass and its error based on a cosh correlation function.
        A lattice of even size is assumed.

        cor: (L/2)
            value of wall to wall correlation function on the first half of the lattice
        cor_err: (L/2)
            error of correlation function on the first half of the lattice

        Returns
        m_eff: (L/2,)
            effective mass
        m_eff_err: (L/2)
            error of the effective mass
        '''
        rel_err = cor_err / cor # relative error
        cor_1, cor_err_1 = np.roll(cor, -1), np.roll(cor_err, -1) # shift to d+1
        rel_err_1 = cor_err_1 / cor_1
        cor__1, cor_err__1 = np.roll(cor, 1), np.roll(cor_err, 1) # shift to d-1
        rel_err__1 = cor_err__1 / cor__1

        A, B = cor_1/cor, cor__1/cor
        x = (A+B)/2 
        m_eff = np.arccosh(x)

        delta_x = 1/2 * (A*(rel_err_1 - rel_err) + B*(rel_err__1 - rel_err))
        # delta_x = A/2*(np.sqrt(rel_err_1**2 + rel_err**2)) + B/2*(np.sqrt(rel_err__1**2 + rel_err**2))
        m_eff_err = np.abs(1/np.sqrt(x**2-1) * delta_x)

        return m_eff, m_eff_err

    data = np.load(data_path)
    ds, cor, cor_err = data[:3]
    m_eff, m_eff_err = effective_mass(cor, cor_err)

    fig = plt.figure()

    plt.errorbar(ds, m_eff, yerr=m_eff_err, fmt='.')
    if xright is not None:
        plt.xlim(right=xright)
    if ybottom is not None:
        plt.ylim(bottom=ybottom)
    if ytop is not None:
        plt.ylim(top=ytop)
    plt.xlabel(r'wall separation [$a$]')
    plt.ylabel(r'effective mass $m_{eff}$')
    fig.gca().xaxis.set_major_locator(MaxNLocator(integer=True)) # set major ticks at integer positions only
    
    if show_plot:
        plt.show()
    # create directories in plot_path if necessary
    dir_path = os.path.dirname(plot_path)
    if dir_path != '':
        os.makedirs(dir_path, exist_ok=True)
    fig.savefig(plot_path)
    plt.close()